:- [operators, sequents, gensym_simple].

dissolve(Formula) :-
	remove_sequent(dissolve),
	dissolve(Formula, Result),
	include(nott Result),
	print_db,
	clear_included. %clear


dissolve(Formula, Formula) :-
	Formula =.. [_].

dissolve(Formula, Result) :-
	Formula =.. [Operator, Right],
	dissolve(Right, RightResult),
	include_if_needed(Operator,
			    RightResult,
			    Result).

dissolve(Formula, Result) :-
	Formula =.. [Operator, Left, Right],
	dissolve(Left, LeftResult),
	dissolve(Right, RightResult),
	include_if_needed(Operator,
			    LeftResult,
			    RightResult,
			    Result).

new_symbol(Simbolis) :-
	gensym(n, Simbolis), !.


include_if_needed(Operator, Right, Result) :-
	Formula =.. [Operator, Right],
	is_included(Formula, Result).

include_if_needed(Operator, Right, Result) :-
	reduce_include(Operator, Right, Result).

include_if_needed(Operator, Left, Right, Result) :-
	Formula =.. [Operator, Left, Right],
	is_included(Formula, Result).

include_if_needed(Operator, Left, Right, Result) :-
	reduce_include(Operator, Left, Right, Result).

include(Formula) :-
	recorda(dissolve, Formula).

mark_included(Formula, NewSymbol) :-
	recorda(included, included(Formula, NewSymbol)).

is_included(Formula, Result) :-
	recorded(included, included(Formula, Result)).

reduce_include(nott, Right, Result) :-
	new_symbol(Result),
% 	include(always(nott Result or nott Right)),%
	( Result @=< Right ->
	    include(always(Result or Right));
	    include(always(Right or Result))
	),
	mark_included(nott Right, Result).

reduce_include(always, Right, Result) :-
	new_symbol(Result),
% 	include(always(nott Result or always Right)),%
	include(always(Result or eventually nott Right)),
	mark_included(always Right, Result).

reduce_include(eventually, Right, Result) :-
	new_symbol(Result),
% 	include(always(nott Result or eventually Right)),%
	include(always(Result or always nott Right)),
	mark_included(eventually Right, Result).

reduce_include(or, Left, Right, Result) :-
	new_symbol(Result),
% 	include(always(nott Result or Left or Right)),%
	( Result @=< nott Left ->
	    include(always(Result or nott Left));
	    include(always(nott Left or Result))
	),
	( Result @=< nott Right ->
	    include(always(Result or nott Right));
	    include(always(nott Right or Result))
	),
	mark_included(Left or Right, Result). %do we have to check which one is more in the left?

reduce_include(and, Left, Right, Result) :-
	new_symbol(Result),
% 	include(always(nott Result or Left)),%
% 	include(always(nott Result or Right)),%
	sort_or(Result, nott Left, nott Right, SortedOr),
	include(always(SortedOr)),
	mark_included(Left and Right, Result).


sort_or(First, Second, Third, First or Second or Third) :-
	First @=< Second,
	Second @=< Third.
sort_or(First, Second, Third, First or Third or Second) :-
	First @=< Third,
	Third @=< Second.
sort_or(First, Second, Third, Second or First or Third) :-
	Second @=< First,
	First @=< Third.
sort_or(First, Second, Third, Second or Third or First) :-
	Second @=< Third,
	Third @=< First.
sort_or(First, Second, Third, Third or First or Second) :-
	Third @=< First,
	First @=< Second.
sort_or(First, Second, Third, Third or Second or First) :-
	Third @=< Second,
	Second @=< First.


print_db :-
	writeln('Initial sequent:'),
	recorded(dissolve, Formula),
	writeln(Formula),
	fail.

print_db :-
	writeln('').

print_included :-
	writeln('Included:'),
	recorded(included, Formula),
	writeln(Formula),
	fail.

print_included :-
	writeln('').

clear_db :-
	recorded(dissolve, _, Reference),
	erase(Reference),
	fail.

clear_db.

clear_included :-
	recorded(included, _, Reference),
	erase(Reference),
	fail.

clear_included.
